from typing import Annotated, List

from marker.processors import BaseProcessor
from marker.schema import BlockTypes
from marker.schema.blocks import Block
from marker.schema.document import Document
from marker.schema.text import Line
from marker.util import matrix_intersection_area


class LineMergeProcessor(BaseProcessor):
    """
    A processor for merging inline math lines.
    """
    block_types = (BlockTypes.Text, BlockTypes.TextInlineMath, BlockTypes.Caption, BlockTypes.Footnote, BlockTypes.SectionHeader)
    min_merge_pct: Annotated[
        float,
        "The minimum percentage of intersection area to consider merging."
    ] = .015
    block_expand_threshold: Annotated[
        float,
        "The percentage of the block width to expand the bounding box."
    ] = .05
    min_merge_ydist: Annotated[
        float,
        "The minimum y distance between lines to consider merging."
    ] = 5
    intersection_pct_threshold: Annotated[
        float,
        "The total amount of intersection area concentrated in the max intersection block."
    ] = .5
    vertical_overlap_pct_threshold: Annotated[
        float,
        "The minimum percentage of vertical overlap to consider merging."
    ] = .8
    use_llm: Annotated[
        bool,
        "Whether to use LLMs to improve accuracy."
    ] = False

    def __init__(self, config):
        super().__init__(config)

    def merge_lines(self, lines: List[Line], block: Block):
        lines = [l for l in lines if l.polygon.width * 5 > l.polygon.height]  # Skip vertical lines
        line_bboxes = [l.polygon.expand(self.block_expand_threshold, 0).bbox for l in lines]  # Expand horizontally
        intersections = matrix_intersection_area(line_bboxes, line_bboxes)

        merges = []
        merge = []
        for i in range(len(line_bboxes)):
            intersection_row = intersections[i]
            intersection_row[i] = 0  # Zero out the current idx

            if i < len(line_bboxes) - 1:
                intersection_row[i+1] = 0 # Zero out the next idx, so we only evaluate merge from the left

            if len(merge) == 0:
                merge.append(i)
                continue

            # Zero out previous merge segments
            merge_intersection = sum([intersection_row[m] for m in merge])
            line_area = lines[i].polygon.area
            intersection_pct = merge_intersection / max(1, line_area)

            total_intersection = max(1, sum(intersection_row))

            line_start = lines[merge[0]].polygon.y_start
            line_end = lines[merge[0]].polygon.y_end

            vertical_overlap_start = max(line_start, lines[i].polygon.y_start)
            vertical_overlap_end = min(line_end, lines[i].polygon.y_end)
            vertical_overlap = max(0, vertical_overlap_end - vertical_overlap_start)
            vertical_overlap_pct = vertical_overlap / max(1, lines[i].polygon.height)

            if all([
                # Overlaps enough
                intersection_pct >= self.min_merge_pct,
                # Within same line
                vertical_overlap_pct > self.vertical_overlap_pct_threshold,
                # doesn't overlap with anything else
                merge_intersection / total_intersection > self.intersection_pct_threshold
            ]):
                merge.append(i)
            else:
                merges.append(merge)
                merge = []

        if merge:
            merges.append(merge)

        merges = [m for m in merges if len(m) > 1]
        merged = set()
        for merge in merges:
            merge = [m for m in merge if m not in merged]
            if len(merge) < 2:
                continue

            line: Line = lines[merge[0]]
            merged.add(merge[0])
            for idx in merge[1:]:
                other_line: Line = lines[idx]
                line.merge(other_line)
                block.structure.remove(other_line.id)
                other_line.removed = True  # Mark line as removed
                merged.add(idx)

            # It is probably math if we are merging provider lines like this
            if not line.formats:
                line.formats = ["math"]
            elif "math" not in line.formats:
                line.formats.append("math")


    def __call__(self, document: Document):
        # Merging lines only needed for inline math
        if not self.use_llm:
            return

        for page in document.pages:
            for block in page.contained_blocks(document, self.block_types):
                if block.structure is None:
                    continue

                if not len(block.structure) >= 2:  # Skip single lines
                    continue

                lines = block.contained_blocks(document, (BlockTypes.Line,))
                self.merge_lines(lines, block)



from marker.schema import BlockTypes
from marker.schema.blocks import Block


class Handwriting(Block):
    block_type: BlockTypes = BlockTypes.Handwriting
    block_description: str = "A region that contains handwriting."
    html: str | None = None
    replace_output_newlines: bool = True

    def assemble_html(self, document, child_blocks, parent_structure):
        if self.html:
            return self.html
        else:
            return super().assemble_html(document, child_blocks, parent_structure)



from collections import defaultdict
from copy import deepcopy
from itertools import chain
from typing import Annotated, List, Optional, Tuple

import numpy as np
from ftfy import fix_text
from PIL import Image

from surya.detection import DetectionPredictor, InlineDetectionPredictor, TextDetectionResult
from surya.ocr_error import OCRErrorPredictor

from marker.builders import BaseBuilder
from marker.providers import ProviderOutput, ProviderPageLines
from marker.providers.pdf import PdfProvider
from marker.schema import BlockTypes
from marker.schema.document import Document
from marker.schema.groups.page import PageGroup
from marker.schema.polygon import PolygonBox
from marker.schema.registry import get_block_class
from marker.schema.text.line import Line
from marker.settings import settings
from marker.util import matrix_intersection_area, sort_text_lines


class TextBox(PolygonBox):
    math: bool = False

    def __hash__(self):
        return hash(tuple(self.bbox))

class LineBuilder(BaseBuilder):
    """
    A builder for detecting text lines, and inline math. Merges the detected lines with the lines from the provider
    """
    detection_batch_size: Annotated[
        Optional[int],
        "The batch size to use for the detection model.",
        "Default is None, which will use the default batch size for the model."
    ] = None
    ocr_error_batch_size: Annotated[
        Optional[int],
        "The batch size to use for the ocr error detection model.",
        "Default is None, which will use the default batch size for the model."
    ] = None
    enable_table_ocr: Annotated[
        bool,
        "Whether to skip OCR on tables.  The TableProcessor will re-OCR them.  Only enable if the TableProcessor is not running.",
    ] = False
    layout_coverage_min_lines: Annotated[
        int,
        "The minimum number of PdfProvider lines that must be covered by the layout model",
        "to consider the lines from the PdfProvider valid.",
    ] = 1
    layout_coverage_threshold: Annotated[
        float,
        "The minimum coverage ratio required for the layout model to consider",
        "the lines from the PdfProvider valid.",
    ] = .25
    min_document_ocr_threshold: Annotated[
        float,
        "If less pages than this threshold are good, OCR will happen in the document.  Otherwise it will not."
    ] = 0.85
    span_inline_math_overlap_threshold: Annotated[
        float,
        "The minimum overlap of a span with an inline math box to consider for removal"
    ] = .5
    char_inline_math_overlap_threshold: Annotated[
        float,
        "The minimum overlap of a character with an inline math box to consider for removal"
    ] = .5
    line_inline_math_overlap_threshold: Annotated[
        float,
        "The minimum overlap of a line with an inline math box to consider as a match"
    ] = 0.
    line_inline_min_overlap_pct: Annotated[
        float,
        "The percentage of a provider line that has to be covered by a math line."
    ] = .3
    line_text_overlap_threshold: Annotated[
        float,
        "The minimum overlap of an equation with a text line to consider as a match"
    ] = .5
    inline_math_line_vertical_merge_threshold: Annotated[
        int,
        "The maximum pixel distance between y1s for two lines to be merged"
    ] = 8
    excluded_for_coverage: Annotated[
        Tuple[BlockTypes],
        "A list of block types to exclude from the layout coverage check.",
    ] = (BlockTypes.Figure, BlockTypes.Picture, BlockTypes.Table, BlockTypes.FigureGroup, BlockTypes.TableGroup, BlockTypes.PictureGroup)
    use_llm: Annotated[
        bool,
        "Whether to use the LLM model for advanced processing."
    ] = False
    texify_inline_spans: Annotated[
        bool,
        "Whether to run texify on inline math spans."
    ] = False
    ocr_remove_blocks: Tuple[BlockTypes, ...] = (BlockTypes.Table, BlockTypes.Form, BlockTypes.TableOfContents, BlockTypes.Equation)
    disable_tqdm: Annotated[
        bool,
        "Disable tqdm progress bars.",
    ] = False

    def __init__(self, detection_model: DetectionPredictor, inline_detection_model: InlineDetectionPredictor, ocr_error_model: OCRErrorPredictor, config=None):
        super().__init__(config)

        self.detection_model = detection_model
        self.inline_detection_model = inline_detection_model
        self.ocr_error_model = ocr_error_model

    def __call__(self, document: Document, provider: PdfProvider):
        # Disable inline detection for documents where layout model doesn't detect any equations
        # Also disable if we won't use the inline detections (if we aren't using the LLM or texify)
        do_inline_math_detection = all([
            len(document.contained_blocks([BlockTypes.Equation, BlockTypes.TextInlineMath])) > 0,
            (self.texify_inline_spans or self.use_llm)
        ])

        provider_lines, ocr_lines = self.get_all_lines(document, provider, do_inline_math_detection)
        self.merge_blocks(document, provider_lines, ocr_lines)

    def get_detection_batch_size(self):
        if self.detection_batch_size is not None:
            return self.detection_batch_size
        elif settings.TORCH_DEVICE_MODEL == "cuda":
            return 4
        return 4

    def get_ocr_error_batch_size(self):
        if self.ocr_error_batch_size is not None:
            return self.ocr_error_batch_size
        elif settings.TORCH_DEVICE_MODEL == "cuda":
            return 4
        return 4

    def get_detection_results(self, page_images: List[Image.Image], run_detection: List[bool], do_inline_math_detection: bool):
        self.detection_model.disable_tqdm = self.disable_tqdm
        page_detection_results = self.detection_model(
            images=page_images,
            batch_size=self.get_detection_batch_size()
        )
        inline_detection_results = [None] * len(page_detection_results)
        if do_inline_math_detection:
            self.inline_detection_model.disable_tqdm = self.disable_tqdm
            inline_detection_results = self.inline_detection_model(
                images=page_images,
                text_boxes=[[b.bbox for b in det_result.bboxes] for det_result in page_detection_results],
                batch_size=self.get_detection_batch_size()
            )

        assert len(page_detection_results) == len(inline_detection_results) == sum(run_detection)
        detection_results = []
        inline_results = []
        idx = 0
        for good in run_detection:
            if good:
                detection_results.append(page_detection_results[idx])
                inline_results.append(inline_detection_results[idx])
                idx += 1
            else:
                detection_results.append(None)
                inline_results.append(None)
        assert idx == len(page_images)

        assert len(run_detection) == len(detection_results) == len(inline_results)
        return detection_results, inline_results


    def get_all_lines(self, document: Document, provider: PdfProvider, do_inline_math_detection: bool):
        ocr_error_detection_results = self.ocr_error_detection(document.pages, provider.page_lines)

        boxes_to_ocr = {page.page_id: [] for page in document.pages}
        page_lines = {page.page_id: [] for page in document.pages}

        LineClass: Line = get_block_class(BlockTypes.Line)

        layout_good = []
        for document_page, ocr_error_detection_label in zip(document.pages, ocr_error_detection_results.labels):
            provider_lines: List[ProviderOutput] = provider.page_lines.get(document_page.page_id, [])
            provider_lines_good = all([
                bool(provider_lines),
                ocr_error_detection_label != 'bad',
                self.check_layout_coverage(document_page, provider_lines)
            ])
            layout_good.append(provider_lines_good)

        # Don't OCR if only a few pages are bad
        if sum(layout_good) > len(document.pages) * self.min_document_ocr_threshold:
            layout_good = [True] * len(document.pages)

        run_detection = [(not good or do_inline_math_detection) for good in layout_good]
        page_images = [page.get_image(highres=False, remove_blocks=self.ocr_remove_blocks) for page, good in zip(document.pages, run_detection) if good]

        # Note: run_detection is longer than page_images, since it has a value for each page, not just good ones
        # Detection results and inline detection results are for every page (we use run_detection to make the list full length)
        detection_results, inline_detection_results = self.get_detection_results(page_images, run_detection, do_inline_math_detection)

        assert len(detection_results) == len(inline_detection_results) == len(layout_good) == len(document.pages)
        for document_page, detection_result, inline_detection_result, provider_lines_good in zip(
                document.pages,
                detection_results,
                inline_detection_results,
                layout_good
        ):
            provider_lines: List[ProviderOutput] = provider.page_lines.get(document_page.page_id, [])
            page_size = provider.get_page_bbox(document_page.page_id).size
            image_size = PolygonBox.from_bbox(detection_result.image_bbox).size if detection_result else page_size

            # Merge text and inline math detection results
            merged_detection_boxes = self.determine_math_lines(text_result=detection_result, inline_result=inline_detection_result)
            # Sort the lines to ensure that the order is preserved
            merged_detection_boxes = sort_text_lines(merged_detection_boxes)

            math_detection_boxes = [(i, box) for i, box in enumerate(merged_detection_boxes) if box.math]
            nonmath_detection_boxes = [(i, box) for i, box in enumerate(merged_detection_boxes) if not box.math]

            if provider_lines_good:
                # Merge inline math blocks into the provider lines, only persist new detected text lines which do not overlap with existing provider lines
                # The missing lines are not from a table, so we can safely set this - The attribute for individual blocks is overridden by OCRBuilder
                document_page.text_extraction_method = 'pdftext'

                # Add in the provider lines - merge ones that get broken by inline math
                page_lines[document_page.page_id].extend(
                    self.merge_provider_lines_inline_math(
                        provider_lines,
                        merged_detection_boxes,
                        image_size,
                        page_size
                    )
                )
            else:
                document_page.text_extraction_method = 'surya'

                # Sort lines properly
                full_lines = nonmath_detection_boxes + math_detection_boxes
                full_lines = sorted(full_lines, key=lambda x: x[0])
                full_lines = [b for _, b in full_lines]

                # Skip inline math merging if no provider lines are good; OCR all text lines and all inline math lines
                boxes_to_ocr[document_page.page_id].extend(full_lines)

        # Dummy lines to merge into the document - Contains no spans, will be filled in later by OCRBuilder
        ocr_lines = {document_page.page_id: [] for document_page in document.pages}
        for page_id, page_ocr_boxes in boxes_to_ocr.items():
            page_size = provider.get_page_bbox(page_id).size
            image_size = document.get_page(page_id).get_image(highres=False).size
            for box_to_ocr in page_ocr_boxes:
                line_polygon = PolygonBox(polygon=box_to_ocr.polygon).rescale(image_size, page_size)
                format = ["math"] if box_to_ocr.math else None
                ocr_lines[page_id].append(
                    ProviderOutput(
                        line=LineClass(
                            polygon=line_polygon,
                            page_id=page_id,
                            text_extraction_method='surya',
                            formats=format
                        ),
                        spans=[]
                    )
                )

        return page_lines, ocr_lines

    def ocr_error_detection(self, pages:List[PageGroup], provider_page_lines: ProviderPageLines):
        page_texts = []
        for document_page in pages:
            provider_lines = provider_page_lines.get(document_page.page_id, [])
            page_text = '\n'.join(' '.join(s.text for s in line.spans) for line in provider_lines)
            page_texts.append(page_text)

        self.ocr_error_model.disable_tqdm = self.disable_tqdm
        ocr_error_detection_results = self.ocr_error_model(
            page_texts,
            batch_size=int(self.get_ocr_error_batch_size())
        )
        return ocr_error_detection_results

    def check_layout_coverage(
        self,
        document_page: PageGroup,
        provider_lines: List[ProviderOutput],
    ):
        covered_blocks = 0
        total_blocks = 0
        large_text_blocks = 0

        layout_blocks = [document_page.get_block(block) for block in document_page.structure]
        layout_blocks = [b for b in layout_blocks if b.block_type not in self.excluded_for_coverage]

        layout_bboxes = [block.polygon.bbox for block in layout_blocks]
        provider_bboxes = [line.line.polygon.bbox for line in provider_lines]

        if len(layout_bboxes) == 0:
            return True

        if len(provider_bboxes) == 0:
            return False

        intersection_matrix = matrix_intersection_area(layout_bboxes, provider_bboxes)

        for idx, layout_block in enumerate(layout_blocks):
            total_blocks += 1
            intersecting_lines = np.count_nonzero(intersection_matrix[idx] > 0)

            if intersecting_lines >= self.layout_coverage_min_lines:
                covered_blocks += 1

            if layout_block.polygon.intersection_pct(document_page.polygon) > 0.8 and layout_block.block_type == BlockTypes.Text:
                large_text_blocks += 1

        coverage_ratio = covered_blocks / total_blocks if total_blocks > 0 else 1
        text_okay = coverage_ratio >= self.layout_coverage_threshold

        # Model will sometimes say there is a single block of text on the page when it is blank
        if not text_okay and (total_blocks == 1 and large_text_blocks == 1):
            text_okay = True
        return text_okay

    def merge_blocks(self, document: Document, page_provider_lines: ProviderPageLines, page_ocr_lines: ProviderPageLines):
        for document_page in document.pages:
            provider_lines = page_provider_lines[document_page.page_id]
            ocr_lines = page_ocr_lines[document_page.page_id]

            # Only one or the other will have lines
            merged_lines = provider_lines + ocr_lines

            # Text extraction method is overridden later for OCRed documents
            document_page.merge_blocks(merged_lines, text_extraction_method='pdftext')


    def determine_math_lines(
        self,
        text_result: TextDetectionResult,
        inline_result: TextDetectionResult,
    ) -> List[TextBox]:
        """
        Marks lines as math if they contain inline math boxes.
        """

        if not text_result:
            return []

        text_boxes = [
            TextBox(
                polygon=box.polygon
            ) for box in text_result.bboxes
        ]
        
        # Skip if no inline math was detected
        if not inline_result:
            return text_boxes

        inline_bboxes = [m.bbox for m in inline_result.bboxes]
        text_bboxes = [t.bbox for t in text_boxes]

        if len(inline_bboxes) == 0:
            return text_boxes

        if len(text_boxes) == 0:
            return []

        overlaps = matrix_intersection_area(inline_bboxes, text_bboxes)

        # Mark text boxes as math if they overlap with an inline math box
        for i, inline_box in enumerate(inline_result.bboxes):
            overlap_row = overlaps[i]
            max_overlap_idx = np.argmax(overlap_row)
            max_overlap_box = text_boxes[max_overlap_idx]

            max_overlap = np.max(overlap_row) / inline_box.area

            # Avoid small or nonoverlapping inline math regions
            if max_overlap <= self.line_inline_math_overlap_threshold:
                continue

            # Ignore vertical lines
            if max_overlap_box.height > max_overlap_box.width * 2:
                continue

            max_overlap_box.math = True

        return text_boxes

    # Add appropriate formats to math spans added by inline math detection
    def add_math_span_format(self, provider_line):
        if not provider_line.line.formats:
            provider_line.line.formats = ["math"]
        elif "math" not in provider_line.line.formats:
            provider_line.line.formats.append("math")

    def merge_provider_lines_inline_math(
        self,
        provider_lines: List[ProviderOutput],
        text_lines: List[TextBox],
        image_size,
        page_size
    ):
        # When provider lines is empty or no inline math detected, return provider lines
        if not provider_lines or not text_lines:
            return provider_lines

        horizontal_provider_lines = [
            (j, provider_line) for j, provider_line in enumerate(provider_lines)
            if provider_line.line.polygon.height < provider_line.line.polygon.width * 5 # Multiply to account for small blocks inside equations, but filter out big vertical lines
        ]
        provider_line_boxes = [p.line.polygon.bbox for _, p in horizontal_provider_lines]
        math_line_boxes = [PolygonBox(polygon=m.polygon).rescale(image_size, page_size).bbox for m in text_lines]

        overlaps = matrix_intersection_area(provider_line_boxes, math_line_boxes)

        # Find potential merges
        merge_lines = defaultdict(list)
        for i in range(len(provider_line_boxes)):
            max_overlap_pct = np.max(overlaps[i]) / max(1, horizontal_provider_lines[i][1].line.polygon.area)
            if max_overlap_pct <= self.line_inline_min_overlap_pct:
                continue

            best_overlap = np.argmax(overlaps[i])
            merge_lines[best_overlap].append(i)

        # Filter to get rid of detected lines that include multiple provider lines
        filtered_merge_lines = defaultdict(list)
        for line_idx in merge_lines:
            merge_segment = []
            prev_line = None
            for ml in merge_lines[line_idx]:
                line = horizontal_provider_lines[ml][1].line.polygon
                if prev_line:
                    close = (
                        abs(line.y_start - prev_line.y_start) < self.inline_math_line_vertical_merge_threshold
                        or
                        abs(line.y_end - prev_line.y_end) < self.inline_math_line_vertical_merge_threshold
                    )
                else:
                    # First line
                    close = True

                prev_line = line
                if close:
                    merge_segment.append(ml)
                else:
                    if merge_segment:
                        filtered_merge_lines[line_idx].append(merge_segment)
                    merge_segment = [ml]
            if merge_segment:
                filtered_merge_lines[line_idx].append(merge_segment)

        # Handle the merging
        already_merged = set()
        potential_merges = []
        for line_idx in filtered_merge_lines:
            potential_merges.extend(chain.from_iterable(filtered_merge_lines[line_idx]))
        potential_merges = set(potential_merges)
        out_provider_lines = [(i, p) for i, p in enumerate(provider_lines) if i not in potential_merges]
        for line_idx in filtered_merge_lines:
            text_line = text_lines[line_idx]
            for merge_section in filtered_merge_lines[line_idx]:
                merge_section = [m for m in merge_section if m not in already_merged]
                if len(merge_section) == 0:
                    continue
                elif len(merge_section) == 1:
                    provider_idx = merge_section[0]
                    merged_line = provider_lines[provider_idx]
                    # Only add math format to single lines if the detected line is math
                    if text_line.math:
                        self.add_math_span_format(merged_line)
                    out_provider_lines.append((provider_idx, merged_line))
                    already_merged.add(merge_section[0])
                else:
                    merge_section = sorted(merge_section)
                    merged_line = None
                    min_idx = min(merge_section)
                    for idx in merge_section:
                        provider_line = deepcopy(provider_lines[idx])
                        if merged_line is None:
                            merged_line = provider_line
                        else:
                            # Combine the spans of the provider line with the merged line
                            merged_line = merged_line.merge(provider_line)
                            # Add math regardless, since we assume heavily broken lines are math lines
                            self.add_math_span_format(merged_line)
                        already_merged.add(idx) # Prevent double merging
                    out_provider_lines.append((min_idx, merged_line))

        # Sort to preserve original order
        out_provider_lines = sorted(out_provider_lines, key=lambda x: x[0])
        out_provider_lines = [p for _, p in out_provider_lines]
        return out_provider_lines

    def clear_line_text(self, provider_line):
        for span in provider_line.spans:
            span.text = ""

    def find_overlapping_math_chars(self, provider_line, math_line_polygon, remove_chars=False):
        # Identify if a character in the provider line overlaps with the inline math line - meaning that the line can be treated as math
        spans = provider_line.spans
        math_overlaps = False

        # For providers which do not surface characters
        if provider_line.chars is None:
            for span in spans:
                if span.polygon.intersection_pct(math_line_polygon) > self.span_inline_math_overlap_threshold:
                    math_overlaps = True
            return math_overlaps

        # For providers which surface characters - find line overlap based on characters
        assert len(spans) == len(provider_line.chars), "Number of spans and characters in provider line do not match"
        for span, span_chars in zip(spans, provider_line.chars):
            if len(span_chars) == 0:
                continue

            char_intersections_areas = matrix_intersection_area([char.polygon.bbox for char in span_chars], [math_line_polygon.bbox]).max(axis=-1)
            char_intersections = char_intersections_areas / np.array([char.polygon.area for char in span_chars])

            new_span_chars = []
            span_overlaps = False
            for char, intersection_pct in zip(span_chars, char_intersections):
                if intersection_pct >= self.char_inline_math_overlap_threshold:
                    span_overlaps = True
                else:
                    new_span_chars.append(char)

            # Remove stray characters that overlap with math lines
            if span_overlaps and remove_chars:
                span.text = fix_text(''.join(c.char for c in new_span_chars))

            math_overlaps = math_overlaps or span_overlaps

        return math_overlaps


from typing import List

from marker.schema import BlockTypes
from marker.schema.blocks.basetable import BaseTable


class Form(BaseTable):
    block_type: BlockTypes = BlockTypes.Form
    block_description: str = "A form, such as a tax form, that contains fields and labels.  It most likely doesn't have a table structure."



import re
from collections import defaultdict
from typing import Annotated, Tuple

import regex
from bs4 import NavigableString
from markdownify import MarkdownConverter
from pydantic import BaseModel

from marker.renderers.html import HTMLRenderer
from marker.schema import BlockTypes
from marker.schema.document import Document


def escape_dollars(text):
    return text.replace("$", r"\$")

def cleanup_text(full_text):
    full_text = re.sub(r'\n{3,}', '\n\n', full_text)
    full_text = re.sub(r'(\n\s){3,}', '\n\n', full_text)
    return full_text.strip()

def get_formatted_table_text(element):

    text = []
    for content in element.contents:
        if content is None:
            continue

        if isinstance(content, NavigableString):
            stripped = content.strip()
            if stripped:
                text.append(escape_dollars(stripped))
        elif content.name == 'br':
            text.append('<br>')
        elif content.name == "math":
            text.append("$" + content.text + "$")
        else:
            content_str = escape_dollars(str(content))
            text.append(content_str)

    full_text = ""
    for i, t in enumerate(text):
        if t == '<br>':
            full_text += t
        elif i > 0 and text[i - 1] != '<br>':
            full_text += " " + t
        else:
            full_text += t
    return full_text


class Markdownify(MarkdownConverter):
    def __init__(self, paginate_output, page_separator, inline_math_delimiters, block_math_delimiters, **kwargs):
        super().__init__(**kwargs)
        self.paginate_output = paginate_output
        self.page_separator = page_separator
        self.inline_math_delimiters = inline_math_delimiters
        self.block_math_delimiters = block_math_delimiters

    def convert_div(self, el, text, convert_as_inline):
        is_page = el.has_attr('class') and el['class'][0] == 'page'
        if self.paginate_output and is_page:
            page_id = el['data-page-id']
            pagination_item = "\n\n" + "{" + str(page_id) + "}" + self.page_separator + "\n\n"
            return pagination_item + text
        else:
            return text

    def convert_p(self, el, text, convert_as_inline):
        hyphens = r'-—¬'
        has_continuation = el.has_attr('class') and 'has-continuation' in el['class']
        if has_continuation:
            block_type = BlockTypes[el['block-type']]
            if block_type in [BlockTypes.TextInlineMath, BlockTypes.Text]:
                if regex.compile(rf'.*[\p{{Ll}}|\d][{hyphens}]\s?$', regex.DOTALL).match(text):  # handle hypenation across pages
                    return regex.split(rf"[{hyphens}]\s?$", text)[0]
                return f"{text} "
            if block_type == BlockTypes.ListGroup:
                return f"{text}"
        return f"{text}\n\n" if text else ""  # default convert_p behavior

    def convert_math(self, el, text, convert_as_inline):
        block = (el.has_attr('display') and el['display'] == 'block')
        if block:
            return "\n" + self.block_math_delimiters[0] + text + self.block_math_delimiters[1] + "\n"
        else:
            return " " + self.inline_math_delimiters[0] + text + self.inline_math_delimiters[1] + " "


    def convert_table(self, el, text, convert_as_inline):
        total_rows = len(el.find_all('tr'))
        colspans = []
        rowspan_cols = defaultdict(int)
        for i, row in enumerate(el.find_all('tr')):
            row_cols = rowspan_cols[i]
            for cell in row.find_all(['td', 'th']):
                colspan = int(cell.get('colspan', 1))
                row_cols += colspan
                for r in range(int(cell.get('rowspan', 1)) - 1):
                    rowspan_cols[i + r] += colspan # Add the colspan to the next rows, so they get the correct number of columns
            colspans.append(row_cols)
        total_cols = max(colspans) if colspans else 0

        grid = [[None for _ in range(total_cols)] for _ in range(total_rows)]

        for row_idx, tr in enumerate(el.find_all('tr')):
            col_idx = 0
            for cell in tr.find_all(['td', 'th']):
                # Skip filled positions
                while col_idx < total_cols and grid[row_idx][col_idx] is not None:
                    col_idx += 1

                # Fill in grid
                value = get_formatted_table_text(cell).replace("\n", " ").replace("|", " ").strip()
                rowspan = int(cell.get('rowspan', 1))
                colspan = int(cell.get('colspan', 1))

                if col_idx >= total_cols:
                    # Skip this cell if we're out of bounds
                    continue

                for r in range(rowspan):
                    for c in range(colspan):
                        try:
                            if r == 0 and c == 0:
                                grid[row_idx][col_idx] = value
                            else:
                                grid[row_idx + r][col_idx + c] = '' # Empty cell due to rowspan/colspan
                        except IndexError:
                            # Sometimes the colspan/rowspan predictions can overflow
                            print(f"Overflow in columns: {col_idx + c} >= {total_cols} or rows: {row_idx + r} >= {total_rows}")
                            continue

                col_idx += colspan

        markdown_lines = []
        col_widths = [0] * total_cols
        for row in grid:
            for col_idx, cell in enumerate(row):
                if cell is not None:
                    col_widths[col_idx] = max(col_widths[col_idx], len(str(cell)))

        add_header_line = lambda: markdown_lines.append('|' + '|'.join('-' * (width + 2) for width in col_widths) + '|')

        # Generate markdown rows
        added_header = False
        for i, row in enumerate(grid):
            is_empty_line = all(not cell for cell in row)
            if is_empty_line and not added_header:
                # Skip leading blank lines
                continue

            line = []
            for col_idx, cell in enumerate(row):
                if cell is None:
                    cell = ''
                padding = col_widths[col_idx] - len(str(cell))
                line.append(f" {cell}{' ' * padding} ")
            markdown_lines.append('|' + '|'.join(line) + '|')

            if not added_header:
                # Skip empty lines when adding the header row
                add_header_line()
                added_header = True

        # Handle one row tables
        if total_rows == 1:
            add_header_line()

        table_md = '\n'.join(markdown_lines)
        return "\n\n" + table_md + "\n\n"

    def convert_a(self, el, text, convert_as_inline):
        text = self.escape(text)
        # Escape brackets and parentheses in text
        text = re.sub(r"([\[\]()])", r"\\\1", text)
        return super().convert_a(el, text, convert_as_inline)

    def convert_span(self, el, text, convert_as_inline):
        if el.get("id"):
            return f'<span id="{el["id"]}">{text}</span>'
        else:
            return text

    def escape(self, text):
        text = super().escape(text)
        if self.options['escape_dollars']:
            text = text.replace('$', r'\$')
        return text

class MarkdownOutput(BaseModel):
    markdown: str
    images: dict
    metadata: dict


class MarkdownRenderer(HTMLRenderer):
    page_separator: Annotated[str, "The separator to use between pages.", "Default is '-' * 48."] = "-" * 48
    inline_math_delimiters: Annotated[Tuple[str], "The delimiters to use for inline math."] = ("$", "$")
    block_math_delimiters: Annotated[Tuple[str], "The delimiters to use for block math."] = ("$$", "$$")

    @property
    def md_cls(self):
        return Markdownify(
            self.paginate_output,
            self.page_separator,
            heading_style="ATX",
            bullets="-",
            escape_misc=False,
            escape_underscores=True,
            escape_asterisks=True,
            escape_dollars=True,
            sub_symbol="<sub>",
            sup_symbol="<sup>",
            inline_math_delimiters=self.inline_math_delimiters,
            block_math_delimiters=self.block_math_delimiters
        )


    def __call__(self, document: Document) -> MarkdownOutput:
        document_output = document.render()
        full_html, images = self.extract_html(document, document_output)
        markdown = self.md_cls.convert(full_html)
        markdown = cleanup_text(markdown)
        return MarkdownOutput(
            markdown=markdown,
            images=images,
            metadata=self.generate_document_metadata(document, document_output)
        )



from typing import Annotated, Dict, List, Tuple

from pydantic import BaseModel

from marker.renderers import BaseRenderer
from marker.schema import BlockTypes
from marker.schema.blocks import Block, BlockOutput
from marker.schema.document import Document
from marker.schema.registry import get_block_class


class JSONBlockOutput(BaseModel):
    id: str
    block_type: str
    html: str
    polygon: List[List[float]]
    bbox: List[float]
    children: List['JSONBlockOutput'] | None = None
    section_hierarchy: Dict[int, str] | None = None
    images: dict | None = None


class JSONOutput(BaseModel):
    children: List[JSONBlockOutput]
    block_type: str = str(BlockTypes.Document)
    metadata: dict


def reformat_section_hierarchy(section_hierarchy):
    new_section_hierarchy = {}
    for key, value in section_hierarchy.items():
        new_section_hierarchy[key] = str(value)
    return new_section_hierarchy


class JSONRenderer(BaseRenderer):
    """
    A renderer for JSON output.
    """
    image_blocks: Annotated[
        Tuple[BlockTypes],
        "The list of block types to consider as images.",
    ] = (BlockTypes.Picture, BlockTypes.Figure)
    page_blocks: Annotated[
        Tuple[BlockTypes],
        "The list of block types to consider as pages.",
    ] = (BlockTypes.Page,)

    def extract_json(self, document: Document, block_output: BlockOutput):
        cls = get_block_class(block_output.id.block_type)
        if cls.__base__ == Block:
            html, images = self.extract_block_html(document, block_output)
            return JSONBlockOutput(
                html=html,
                polygon=block_output.polygon.polygon,
                bbox=block_output.polygon.bbox,
                id=str(block_output.id),
                block_type=str(block_output.id.block_type),
                images=images,
                section_hierarchy=reformat_section_hierarchy(block_output.section_hierarchy)
            )
        else:
            children = []
            for child in block_output.children:
                child_output = self.extract_json(document, child)
                children.append(child_output)

            return JSONBlockOutput(
                html=block_output.html,
                polygon=block_output.polygon.polygon,
                bbox=block_output.polygon.bbox,
                id=str(block_output.id),
                block_type=str(block_output.id.block_type),
                children=children,
                section_hierarchy=reformat_section_hierarchy(block_output.section_hierarchy)
            )

    def __call__(self, document: Document) -> JSONOutput:
        document_output = document.render()
        json_output = []
        for page_output in document_output.children:
            json_output.append(self.extract_json(document, page_output))
        return JSONOutput(
            children=json_output,
            metadata=self.generate_document_metadata(document, document_output)
        )



from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Annotated, List, Tuple, Literal

from pydantic import BaseModel
from tqdm import tqdm
from PIL import Image

from marker.processors.llm import BaseLLMComplexBlockProcessor
from marker.schema import BlockTypes
from marker.schema.blocks import Block, TableCell
from marker.schema.document import Document


class LLMTableMergeProcessor(BaseLLMComplexBlockProcessor):
    block_types: Annotated[
        Tuple[BlockTypes],
        "The block types to process.",
    ] = (BlockTypes.Table, BlockTypes.TableOfContents)
    table_height_threshold: Annotated[
        float,
        "The minimum height ratio relative to the page for the first table in a pair to be considered for merging.",
    ] = 0.6
    table_start_threshold: Annotated[
        float,
        "The maximum percentage down the page the second table can start to be considered for merging."
    ] = 0.2
    vertical_table_height_threshold: Annotated[
        float,
        "The height tolerance for 2 adjacent tables to be merged into one."
    ] = 0.25
    vertical_table_distance_threshold: Annotated[
        int,
        "The maximum distance between table edges for adjacency."
    ] = 20
    horizontal_table_width_threshold: Annotated[
        float,
        "The width tolerance for 2 adjacent tables to be merged into one."
    ] = 0.25
    horizontal_table_distance_threshold: Annotated[
        int,
        "The maximum distance between table edges for adjacency."
    ] = 10
    column_gap_threshold: Annotated[
        int,
        "The maximum gap between columns to merge tables"
    ] = 50
    disable_tqdm: Annotated[
        bool,
        "Whether to disable the tqdm progress bar.",
    ] = False
    table_merge_prompt: Annotated[
        str,
        "The prompt to use for rewriting text.",
        "Default is a string containing the Gemini rewriting prompt."
    ] = """You're a text correction expert specializing in accurately reproducing tables from PDFs.
You'll receive two images of tables from successive pages of a PDF.  Table 1 is from the first page, and Table 2 is from the second page.  Both tables may actually be part of the same larger table. Your job is to decide if Table 2 should be merged with Table 1, and how they should be joined.  The should only be merged if they're part of the same larger table, and Table 2 cannot be interpreted without merging.

You'll specify your judgement in json format - first whether Table 2 should be merged with Table 1, then the direction of the merge, either `bottom` or `right`.  A bottom merge means that the rows of Table 2 are joined to the rows of Table 1. A right merge means that the columns of Table 2 are joined to the columns of Table 1.  (bottom merge is equal to np.vstack, right merge is equal to np.hstack)

Table 2 should be merged at the bottom of Table 1 if Table 2 has no headers, and the rows have similar values, meaning that Table 2 continues Table 1. Table 2 should be merged to the right of Table 1 if each row in Table 2 matches a row in Table 1, meaning that Table 2 contains additional columns that augment Table 1.

Only merge Table 1 and Table 2 if Table 2 cannot be interpreted without merging.  Only merge Table 1 and Table 2 if you can read both images properly.

**Instructions:**
1. Carefully examine the provided table images.  Table 1 is the first image, and Table 2 is the second image.
2. Examine the provided html representations of Table 1 and Table 2.
3. Write a description of Table 1.
4. Write a description of Table 2.
5. Analyze whether Table 2 should be merged into Table 1, and write an explanation.
6. Output your decision on whether they should be merged, and merge direction.
**Example:**
Input:
Table 1
```html
<table>
    <tr>
        <th>Name</th>
        <th>Age</th>
        <th>City</th>
        <th>State</th>
    </tr>
    <tr>
        <td>John</td>
        <td>25</td>
        <td>Chicago</td>
        <td>IL</td>
    </tr>
```
Table 2
```html
<table>
    <tr>
        <td>Jane</td>
        <td>30</td>
        <td>Los Angeles</td>
        <td>CA</td>
    </tr>
```
Output:
```json
{
    "table1_description": "Table 1 has 4 headers, and 1 row.  The headers are Name, Age, City, and State.",
    "table2_description": "Table 2 has no headers, but the values appear to represent a person's name, age, city, and state.",
    "explanation": "The values in Table 2 match the headers in Table 1, and Table 2 has no headers. Table 2 should be merged to the bottom of Table 1.",
    "merge": "true",
    "direction": "bottom"
}
```
**Input:**
Table 1
```html
{{table1}}
Table 2
```html
{{table2}}
```
"""

    @staticmethod
    def get_row_count(cells: List[TableCell]):
        if not cells:
            return 0

        max_rows = None
        for col_id in set([cell.col_id for cell in cells]):
            col_cells = [cell for cell in cells if cell.col_id == col_id]
            rows = 0
            for cell in col_cells:
                rows += cell.rowspan
            if max_rows is None or rows > max_rows:
                max_rows = rows
        return max_rows

    @staticmethod
    def get_column_count(cells: List[TableCell]):
        if not cells:
            return 0

        max_cols = None
        for row_id in set([cell.row_id for cell in cells]):
            row_cells = [cell for cell in cells if cell.row_id == row_id]
            cols = 0
            for cell in row_cells:
                cols += cell.colspan
            if max_cols is None or cols > max_cols:
                max_cols = cols
        return max_cols

    def rewrite_blocks(self, document: Document):
        pbar = tqdm(desc=f"{self.__class__.__name__} running", disable=self.disable_tqdm)
        table_runs = []
        table_run = []
        prev_block = None
        prev_page_block_count = None
        for page in document.pages:
            page_blocks = page.contained_blocks(document, self.block_types)
            for block in page_blocks:
                merge_condition = False
                if prev_block is not None:
                    prev_cells = prev_block.contained_blocks(document, (BlockTypes.TableCell,))
                    curr_cells = block.contained_blocks(document, (BlockTypes.TableCell,))
                    row_match = abs(self.get_row_count(prev_cells) - self.get_row_count(curr_cells)) < 5, # Similar number of rows
                    col_match = abs(self.get_column_count(prev_cells) - self.get_column_count(curr_cells)) < 2

                    subsequent_page_table = all([
                        prev_block.page_id == block.page_id - 1, # Subsequent pages
                        max(prev_block.polygon.height / page.polygon.height,
                            block.polygon.height / page.polygon.height) > self.table_height_threshold, # Take up most of the page height
                            (len(page_blocks) == 1 or prev_page_block_count == 1), # Only table on the page
                            (row_match or col_match)
                        ])

                    same_page_vertical_table = all([
                        prev_block.page_id == block.page_id, # On the same page
                        (1 - self.vertical_table_height_threshold) < prev_block.polygon.height / block.polygon.height < (1 + self.vertical_table_height_threshold), # Similar height
                        abs(block.polygon.x_start - prev_block.polygon.x_end) < self.vertical_table_distance_threshold, # Close together in x
                        abs(block.polygon.y_start - prev_block.polygon.y_start) < self.vertical_table_distance_threshold, # Close together in y
                        row_match
                    ])

                    same_page_horizontal_table = all([
                        prev_block.page_id == block.page_id, # On the same page
                        (1 - self.horizontal_table_width_threshold) < prev_block.polygon.width / block.polygon.width < (1 + self.horizontal_table_width_threshold), # Similar width
                        abs(block.polygon.y_start - prev_block.polygon.y_end) < self.horizontal_table_distance_threshold, # Close together in y
                        abs(block.polygon.x_start - prev_block.polygon.x_start) < self.horizontal_table_distance_threshold, # Close together in x
                        col_match
                    ])

                    same_page_new_column = all([
                        prev_block.page_id == block.page_id, # On the same page
                        abs(block.polygon.x_start - prev_block.polygon.x_end) < self.column_gap_threshold,
                        block.polygon.y_start < prev_block.polygon.y_end,
                        block.polygon.width * (1 - self.vertical_table_height_threshold) < prev_block.polygon.width  < block.polygon.width * (1 + self.vertical_table_height_threshold), # Similar width
                        col_match
                    ])
                    merge_condition = any([subsequent_page_table, same_page_vertical_table, same_page_new_column, same_page_horizontal_table])

                if prev_block is not None and merge_condition:
                    if prev_block not in table_run:
                        table_run.append(prev_block)
                    table_run.append(block)
                else:
                    if table_run:
                        table_runs.append(table_run)
                    table_run = []
                prev_block = block
            prev_page_block_count = len(page_blocks)

        if table_run:
            table_runs.append(table_run)

        with ThreadPoolExecutor(max_workers=self.max_concurrency) as executor:
            for future in as_completed([
                executor.submit(self.process_rewriting, document, blocks)
                for blocks in table_runs
            ]):
                future.result()  # Raise exceptions if any occurred
                pbar.update(1)

        pbar.close()

    def process_rewriting(self, document: Document, blocks: List[Block]):
        if len(blocks) < 2:
            # Can't merge single tables
            return

        start_block = blocks[0]
        for i in range(1, len(blocks)):
            curr_block = blocks[i]
            children = start_block.contained_blocks(document, (BlockTypes.TableCell,))
            children_curr = curr_block.contained_blocks(document, (BlockTypes.TableCell,))
            if not children or not children_curr:
                # Happens if table/form processors didn't run
                break

            start_image = start_block.get_image(document, highres=False)
            curr_image = curr_block.get_image(document, highres=False)
            start_html = start_block.render(document).html
            curr_html = curr_block.render(document).html

            prompt = self.table_merge_prompt.replace("{{table1}}", start_html).replace("{{table2}}", curr_html)

            response = self.llm_service(
                prompt,
                [start_image, curr_image],
                curr_block,
                MergeSchema,
            )

            if not response or ("direction" not in response or "merge" not in response):
                curr_block.update_metadata(llm_error_count=1)
                break

            merge = response["merge"]

            # The original table is okay
            if "true" not in merge:
                start_block = curr_block
                continue

            # Merge the cells and images of the tables
            direction = response["direction"]
            if not self.validate_merge(children, children_curr, direction):
                start_block = curr_block
                continue

            merged_image = self.join_images(start_image, curr_image, direction)
            merged_cells = self.join_cells(children, children_curr, direction)
            curr_block.structure = []
            start_block.structure = [b.id for b in merged_cells]
            start_block.lowres_image = merged_image

    def validate_merge(self, cells1: List[TableCell], cells2: List[TableCell], direction: Literal['right', 'bottom'] = 'right'):
        if direction == "right":
            # Check if the number of rows is the same
            cells1_row_count = self.get_row_count(cells1)
            cells2_row_count = self.get_row_count(cells2)
            return abs(cells1_row_count - cells2_row_count) < 5
        elif direction == "bottom":
            # Check if the number of columns is the same
            cells1_col_count = self.get_column_count(cells1)
            cells2_col_count = self.get_column_count(cells2)
            return abs(cells1_col_count - cells2_col_count) < 2


    def join_cells(self, cells1: List[TableCell], cells2: List[TableCell], direction: Literal['right', 'bottom'] = 'right') -> List[TableCell]:
        if direction == 'right':
            # Shift columns right
            col_count = self.get_column_count(cells1)
            for cell in cells2:
                cell.col_id += col_count
            new_cells = cells1 + cells2
        else:
            # Shift rows up
            row_count = self.get_row_count(cells1)
            for cell in cells2:
                cell.row_id += row_count
            new_cells = cells1 + cells2
        return new_cells

    @staticmethod
    def join_images(image1: Image.Image, image2: Image.Image, direction: Literal['right', 'bottom'] = 'right') -> Image.Image:
        # Get dimensions
        w1, h1 = image1.size
        w2, h2 = image2.size

        if direction == 'right':
            new_height = max(h1, h2)
            new_width = w1 + w2
            new_img = Image.new('RGB', (new_width, new_height), 'white')
            new_img.paste(image1, (0, 0))
            new_img.paste(image2, (w1, 0))
        else:
            new_width = max(w1, w2)
            new_height = h1 + h2
            new_img = Image.new('RGB', (new_width, new_height), 'white')
            new_img.paste(image1, (0, 0))
            new_img.paste(image2, (0, h1))
        return new_img


class MergeSchema(BaseModel):
    table1_description: str
    table2_description: str
    explanation: str
    merge: Literal["true", "false"]
    direction: Literal["bottom", "right"]